# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.scan_paths"""

import copy
import gc
import json
import logging
import os
import tarfile
import threading
import time
import uuid

from datetime import datetime
from multiprocessing import cpu_count

import bitmath
import six

from boto3.s3 import transfer as S3Transfer

from sonicprobe import helpers
from sonicprobe.libs import workerpool

from .common import (DATE_FORMAT, STATUS_PROCESSING, STATUS_STOPPED)
from .s3 import (XtrargonTxProgress, XTRARGON_FILE)


LOG                         = logging.getLogger('xtrargon_ng.classes.scan_paths')

DEFAULT_MAX_WORKERS         = 'auto'
DEFAULT_MAX_LIFETIME        = 3600
DEFAULT_MAX_TASKS           = 100
DEFAULT_MAX_UPLOAD_ATTEMPTS = 10
DEFAULT_OBJECT_TTL          = 604800
DEFAULT_DEST_FORMAT         = "/{server_id}/{backup_type}/{time:%Y-%m-%dT%H:%M:%S}"
DEFAULT_STATE_FILE_FORMAT   = DEFAULT_DEST_FORMAT + "/xtrargon.state"
DEFAULT_COMP_MAX_TOTAL_SIZE = "1G"
DEFAULT_COMP_MAX_FILE_SIZE  = "1G"


class BackupsScanPaths(threading.Thread):
    def __init__(self, module):
        threading.Thread.__init__(self)
        self.name              = 'scan'
        self._module           = module
        self._config           = module.config
        self._s3helper         = module.s3helper

        self._step             = 1
        self._killed           = False

        self._bucket           = self._s3helper.get_bucket()
        self._storage_class    = self._s3helper.get_storage_class()
        self._client           = self._s3helper.boto3_client()
        self._object_ttl       = DEFAULT_OBJECT_TTL
        self._file_sync        = module.file_sync

        self._backup_type      = 'full'

        self._wcompress        = None
        self._wupload          = None

        self._files_cached     = {}
        self._files_uploaded   = {}
        self._files_compressed = []

        if 'object_ttl' in self._config['backups']:
            self._object_ttl = int(self._config['backups']['object_ttl'])

        for x in ('compress', 'upload'):
            if not isinstance(self._config['backups'].get(x), dict):
                self._config['backups'][x] = {}

    @staticmethod
    def _load_workers(name, config):
        max_workers     = config.get('max_workers', DEFAULT_MAX_WORKERS) or 1
        if max_workers == 'auto':
            max_workers = cpu_count()
        elif max_workers < 1:
            max_workers = 1

        worker_lifetime = int(config.get('worker_lifetime', DEFAULT_MAX_LIFETIME))
        max_tasks       = int(config.get('max_tasks', DEFAULT_MAX_TASKS))

        return workerpool.WorkerPool(max_workers = int(max_workers),
                                     life_time   = worker_lifetime,
                                     max_tasks   = max_tasks,
                                     name        = name,
                                     auto_gc     = False)

    @staticmethod
    def _mk_sum(filepath):
        if not os.path.exists(filepath):
            return ''

        fstat = os.stat(filepath)

        return '|'.join((str(fstat.st_size),
                         str(fstat.st_mtime)))

    def _mk_filekey(self, filepath, xid = None):
        if xid:
            return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

        return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

    def _chk_file_cached(self, toppath, srcpath, fkey, nsum):
        if toppath not in self._files_cached:
            self._files_cached[toppath] = {}

        if srcpath in self._files_cached[toppath]:
            osum = self._files_cached[toppath][srcpath]
            if osum and osum == nsum:
                return True

        osum = self._file_sync.get_key(fkey)
        if osum and osum == nsum:
            return True

        return False

    def _compress_db_params(self, params):
        r = {'dest_params': None,
             'enabled': False,
             'max_file_size':  bitmath.parse_string_unsafe(DEFAULT_COMP_MAX_FILE_SIZE).to_Byte().value,
             'max_upload_attempts': DEFAULT_MAX_UPLOAD_ATTEMPTS,
             'max_total_size': bitmath.parse_string_unsafe(DEFAULT_COMP_MAX_TOTAL_SIZE).to_Byte().value}

        if not params or not isinstance(params, dict):
            return r

        r['dest_params'] = params.get('dest_params')
        r['enabled'] = params.get('enabled') or False
        r['max_file_size'] = bitmath.parse_string_unsafe(
            params.get('max_file_size', DEFAULT_COMP_MAX_FILE_SIZE)).to_Byte().value
        r['max_upload_attempts'] = self._get_max_upload_attempts(params)
        r['max_total_size'] = bitmath.parse_string_unsafe(
            params.get('max_total_size', DEFAULT_COMP_MAX_TOTAL_SIZE)).to_Byte().value

        return r

    def _s3up(self, src, dst, max_attempts = 1, delete = False):
        if not dst:
            dst = src

        if not os.path.exists(src):
            return (None, src, dst, delete)

        last_exception = None

        for i in range(1, max_attempts + 1):
            try:
                with S3Transfer.S3Transfer(self._s3helper.boto3_client(),
                                           self._s3helper.transfer_config()) as transfer:
                    transfer.upload_file(
                        filename   = src,
                        bucket     = self._bucket,
                        key        = dst.lstrip('/'),
                        callback   = XtrargonTxProgress(src, False, self._killed),
                        extra_args = {'StorageClass': self._storage_class})

                LOG.info("s3 upload (src: %r, dst: %r, bucket: %r, attempt: %s/%s)",
                         src,
                         dst,
                         self._bucket,
                         i,
                         max_attempts)
                return (True, src, dst, delete)
            except GeneratorExit:
                return (GeneratorExit(), src, dst, delete)
            except Exception as e:
                LOG.debug("Retrying exception caught (%s), "
                          "retrying request, (attempt %s/%s)", e, i,
                          max_attempts, exc_info=True)
                last_exception = e
                continue

        if self._step > 1:
            return (Exception(last_exception), src, dst, delete)

        return (None, src, dst, delete)

    def _s3up_complete(self, ret):
        result, srcpath, destpath, delete = ret

        if isinstance(result, GeneratorExit):
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        elif isinstance(result, Exception):
            self._module.add_errors(repr(result))
            LOG.exception(result)
        elif result is True:
            if destpath:
                self._files_uploaded.pop(destpath, None)
            if delete and os.path.exists(srcpath):
                try:
                    os.unlink(srcpath)
                except Exception:
                    pass

    def _s3api(self, cmd, key = None, **kwargs):
        kwargs['Bucket'] = self._bucket
        if key:
            kwargs['Key'] = key.lstrip('/')
        kwargs['StorageClass'] = self._storage_class

        LOG.debug("s3 api: %r: %r", cmd, kwargs)
        getattr(self._client, cmd)(**kwargs)

    def _put_file(self, xpath, body = ''):
        self._s3api('put_object', xpath, **{'Body': body})

    @staticmethod
    def _destpath(toppath, params, srcpath, xvars):
        destpath = srcpath
        fkwargs  = xvars.copy()
        fkwargs['filepath'] = srcpath
        fkwargs['dirname']  = os.path.dirname(srcpath)
        fkwargs['filename'] = os.path.basename(srcpath)

        if params is True:
            return destpath

        if not params or not isinstance(params, dict):
            params = {'format': DEFAULT_DEST_FORMAT,
                      'prefix': None,
                      'subst':  True}

        if params.get('format'):
            destpath = params['format'].format(**fkwargs)

        if params.get('subst'):
            if isinstance(params['subst'], six.string_types):
                xsubst = params['subst']
            else:
                xsubst = toppath

            if srcpath.startswith(xsubst):
                substpath = srcpath[len(xsubst):].strip(os.path.sep)
                if params.get('format'):
                    destpath = os.path.join(os.path.sep, destpath.strip(os.path.sep), substpath)
                else:
                    destpath = os.path.join(os.path.sep, substpath)

        if params.get('prefix'):
            destpath = os.path.join(params['prefix'].strip(os.path.sep),
                                    destpath.strip(os.path.sep))

        return os.path.join(os.path.sep, destpath.strip(os.path.sep))

    @staticmethod
    def _get_max_upload_attempts(params):
        r = int(params.get('max_upload_attempts', DEFAULT_MAX_UPLOAD_ATTEMPTS) or 1)
        if r < 1:
            r = 1

        return r

    def _pre_cache(self, toppath, srcpath, force = False):
        xid  = self._module.get_state_value('started_at')
        fkey = self._mk_filekey(srcpath, xid)
        nsum = self._mk_sum(srcpath)

        if not force and self._chk_file_cached(toppath, srcpath, fkey, nsum):
            return False

        return (fkey, nsum)

    def _post_cache(self, toppath, srcpath, fkey, nsum):
        if toppath not in self._files_cached:
            self._files_cached[toppath] = {}
        self._files_cached[toppath][srcpath] = nsum

        self._file_sync.set_key(fkey, nsum, self._object_ttl)

    def _upload(self, toppath, srcpath, destpath, **opts):
        xopts = {'add_uploaded': True,
                 'cache':        True,
                 'delete':       False,
                 'max_attempts': 1}
        xopts.update(opts)

        res = None

        if xopts['cache']:
            res = self._pre_cache(toppath, srcpath)
            if not res:
                return False

        if xopts['add_uploaded']:
            self._files_uploaded[destpath] = {'toppath': toppath,
                                              'srcpath': srcpath,
                                              'destpath': destpath}

        self._wupload.run(self._s3up,
                          complete     = self._s3up_complete,
                          src          = srcpath,
                          dst          = destpath,
                          max_attempts = xopts['max_attempts'],
                          delete       = xopts['delete'])

        if xopts['cache']:
            self._post_cache(toppath, srcpath, res[0], res[1])

        return True

    def _db_to_compress(self, toppath, params, xvars):
        r = {}

        if not os.path.isdir(toppath):
            return r

        for xdir in os.listdir(toppath):
            if self._killed:
                raise GeneratorExit()

            subdir    = os.path.join(toppath, xdir)

            if xdir in ('.', '..') or not os.path.isdir(subdir):
                continue

            if params['exclude_patterns'] and params['exclude_patterns'](subdir):
                LOG.debug("exclude path from scan. (path: %r)", subdir)
                continue

            destpath  = self._destpath(toppath,
                                       params['compress_db']['dest_params'],
                                       subdir,
                                       xvars)
            r[subdir] = {'comppath': "%s.tar" % destpath,
                         'cnt': 0,
                         'dbpath': subdir,
                         'size': 0}

            for root, dirs, files in os.walk(subdir, topdown = True, followlinks = params['follow']):
                stop_iter = False

                if self._killed:
                    raise GeneratorExit()

                if params['exclude_patterns'] and params['exclude_patterns'](root):
                    dirs[:] = []
                    LOG.debug("exclude path from scan. (path: %r)", root)
                    break

                if not files:
                    continue

                for filename in files:
                    srcpath = os.path.join(root, filename)

                    if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                        LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                        continue

                    filesize = os.path.getsize(srcpath)
                    if filesize >= params['compress_db']['max_file_size']:
                        stop_iter = True
                        break

                    r[root]['size'] += filesize
                    r[root]['cnt']  += 1
                    if r[root]['size'] >= params['compress_db']['max_total_size']:
                        stop_iter = True
                        break

                if stop_iter:
                    del r[subdir]
                    break

            if subdir not in r or r[subdir]['cnt'] == 0:
                LOG.info("database %r isn't compressible", subdir)
                r.pop(subdir, None)

        return r

    def _compress_db(self, comppath, toppath, dbroot, destpath, params):
        LOG.info("writing tar file: %r", comppath)

        max_attempts = params['compress_db']['max_upload_attempts']
        tar          = None

        try:
            helpers.make_dirs(os.path.dirname(comppath))

            with tarfile.open(comppath, 'w') as tar:
                self._files_compressed.append(comppath)

                # pylint: disable-msg=unused-variable
                for root, dirs, files in os.walk(dbroot, topdown = True, followlinks = params['follow']):
                    if self._killed:
                        raise GeneratorExit()

                    for filename in files:
                        if self._killed:
                            raise GeneratorExit()

                        srcpath = os.path.join(root, filename)

                        if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                            LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                            continue

                        res = self._pre_cache(toppath, srcpath)
                        if not res:
                            continue

                        tar.add(srcpath)

                        self._post_cache(toppath, srcpath, res[0], res[1])
        except GeneratorExit:
            return (GeneratorExit(), toppath, comppath, destpath, max_attempts)
        except Exception as e:
            return (Exception(e), toppath, comppath, destpath, max_attempts)
        finally:
            if tar:
                tar.close()

        return (True, toppath, comppath, destpath, max_attempts)

    def _compress_db_complete(self, ret):
        result, toppath, comppath, destpath, max_attempts = ret

        if isinstance(result, GeneratorExit):
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        elif isinstance(result, Exception):
            self._module.add_errors(repr(result))
            LOG.exception(result)
        elif result is True:
            self._upload(toppath,
                         comppath,
                         destpath,
                         cache        = False,
                         delete       = True,
                         max_attempts = max_attempts)

    def _walk_from_cache(self, toppath, params, xvars):
        LOG.info("walk from cache: %r", toppath)

        files_cached = copy.deepcopy(self._files_cached)

        for srcpath in six.iterkeys(files_cached[toppath]):
            if self._killed:
                raise GeneratorExit()

            max_attempts = self._get_max_upload_attempts(params)
            destpath     = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

            self._upload(toppath, srcpath, destpath, max_attempts = max_attempts)

        del files_cached

    def _walk(self, toppath, params, xvars, dbcomp = None):
        LOG.info("walk path: %r", toppath)

        if self._step == 1:
            max_attempts = 1
        else:
            max_attempts = self._get_max_upload_attempts(params)

        for root, dirs, files in os.walk(toppath, topdown = True, followlinks = params['follow']):
            if self._killed:
                raise GeneratorExit()

            if params['exclude_patterns'] and params['exclude_patterns'](root):
                dirs[:] = []
                LOG.debug("exclude path from scan. (path: %r)", root)
                continue

            if not dirs and not files:
                srcpath  = os.path.join(root, XTRARGON_FILE)
                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)
                self._put_file(destpath)
                continue

            if self._step == 1 and self._wcompress and dbcomp and root in dbcomp:
                comppath = dbcomp[root]['comppath']
                srcpath  = os.path.join(dbcomp[root]['dbpath'],
                                        os.path.basename(comppath))
                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

                self._wcompress.run(self._compress_db,
                                    complete = self._compress_db_complete,
                                    comppath = comppath,
                                    toppath  = toppath,
                                    dbroot   = root,
                                    destpath = destpath,
                                    params   = params)
                dirs[:]  = []
                files[:] = []
                continue

            for filename in files:
                if self._killed:
                    raise GeneratorExit()

                srcpath = os.path.join(root, filename)

                if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                    LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                    continue

                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

                self._upload(toppath, srcpath, destpath, max_attempts = max_attempts)

    def _run(self, fkwargs):
        for toppath, params in six.iteritems(self._config['backups']['paths']):
            toppath = os.path.abspath(toppath)
            dbcomp  = None

            if self._killed:
                raise GeneratorExit()

            params['compress_db'] = self._compress_db_params(params.get('compress_db'))

            if self._step == 1 and params['compress_db']['enabled']:
                if not self._wcompress:
                    self._wcompress = self._load_workers('wcompress', self._config['backups']['compress'])

                dbcomp = self._db_to_compress(toppath, params, fkwargs)

            if self._step == 1 \
               or not bool(params.get('second_step_from_cache')) \
               or toppath not in self._files_cached:
                self._walk(toppath, params, fkwargs, dbcomp)
            else:
                self._walk_from_cache(toppath, params, fkwargs)

        dbcomp = None

        if self._wcompress:
            while not self._killed:
                if self._wcompress.count_working() == 0 \
                   and self._wcompress.tasks.empty():
                    break

                time.sleep(0.5)

        if self._wupload:
            while not self._killed:
                if self._wupload.count_working() == 0 \
                   and self._wupload.tasks.empty():
                    break

                time.sleep(0.5)

        while self._files_uploaded:
            if self._killed:
                raise GeneratorExit()

            files_uploaded = copy.deepcopy(self._files_uploaded)

            for key, item in six.iteritems(files_uploaded):
                if self._killed:
                    raise GeneratorExit()

                if not os.path.exists(item['srcpath']):
                    self._files_uploaded.pop(key, None)
                    continue

                if self._upload(item['toppath'],
                                item['srcpath'],
                                item['destpath'],
                                add_uploaded = False):
                    continue

                self._files_uploaded.pop(key, None)

            del files_uploaded

            time.sleep(0.5)

    def _clean_files_compressed(self):
        if not self._files_compressed:
            return

        dirs_comp = {}

        for tar in self._files_compressed:
            dirs_comp[os.path.dirname(tar)] = 1
            if os.path.exists(tar):
                os.unlink(tar)

        for dir_comp in six.iterkeys(dirs_comp):
            try:
                os.rmdir(dir_comp)
            except OSError:
                pass

        del dirs_comp

    def _post_run(self, fkwargs):
        LOG.info("finalizing backup")

        self._files_cached   = {}
        self._files_uploaded = {}

        try:
            state = self._module.get_state(end = True)
            state_file_format = self._config['backups'].get('state_file_format') or DEFAULT_STATE_FILE_FORMAT
            self._put_file(state_file_format.format(**fkwargs),
                           body = json.dumps(state))

            if self._config['backups'].get('notifications'):
                self._module.notifiers(state)

            if self._wcompress:
                self._wcompress.killall(1)
                self._wcompress = None

            if self._wupload:
                self._wupload.killall(1)
                self._wupload = None

            self._clean_files_compressed()

            gc.collect()
        except Exception as e:
            LOG.exception(e)

    def run(self):
        self._module.is_running.wait()
        LOG.info("starting backup")
        self._module.set_state_value('started_at', time.strftime(DATE_FORMAT))
        self._module.set_state_value('uuid', "%s" % uuid.uuid4())
        self._module.set_state_value('status', STATUS_PROCESSING)

        if self._config['backups'].get('notifications'):
            self._module.notifiers(self._module.get_state())

        self._module.is_scanning.set()

        self._files_cached     = {}
        self._files_uploaded   = {}
        self._files_compressed = []

        fkwargs = {'env': os.environ,
                   'backup_type': self._backup_type,
                   'gmtime': datetime.utcnow(),
                   'time': datetime.now(),
                   'server_id': self._module.server_id}

        try:
            self._wupload = self._load_workers('wupload',
                                               self._config['backups']['upload'])

            if self._killed:
                raise GeneratorExit()

            LOG.info("first step of backup")
            self._step = 1
            self._run(fkwargs)
            self._module.lock_db.set()
            self._module.db_locked.wait()

            if self._killed:
                raise GeneratorExit()

            LOG.info("second step of backup")
            self._step = 2
            self._run(fkwargs)
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            self._module.lock_db.clear()
            self._module.is_scanning.clear()
            self._module.is_running.clear()

        self._post_run(fkwargs)

    def terminate(self):
        LOG.info("backup terminated")
        self._killed = True
        self._module.is_running.set()
        self._module.is_scanning.set()

        if self._wcompress:
            self._wcompress.killall(1)

        if self._wupload:
            self._wupload.killall(1)
